static char _rcsid_[] =
"$Header: readMPPE.c,v 1.9 91/10/23 16:36:10 jothy Exp $";
/*******************************************************************
                        Copyright (c) 1991
                   MasPar Computer Corporation (R)

  The terms and conditions of the Software License Agreement
  executed by the parties shall apply to receipt and use of this 
  Source File ("Source"). Those terms are supplemented by the
  following additional terms.

  Source is provided to licensed users of MasPar hardware and software to
  enable Licensee to make modifications to customize, port, or otherwise
  adapt Source for their internal requirements only.  For this purpose,
  Licensee is authorized to make copies of Source, provided Licensee
  ensures that MasPar's copyright notices are included on said copies.
  Any copies made are subject to these and the Software License terms.

  All rights in Source, as well as any future modifications or
  improvements provided by MasPar, shall remain with MasPar and no title
  to the intellectual property is transferred to Licensee.
  
  MasPar shall not be responsible for maintenance of any Source
  modified by Licensee. Further, Licensee agrees to indemnify and
  hold MasPar harmless against all claims arising out of the
  functioning or malfunctioning of Source modified by Licensee.

  MasPar makes no warranty of any kind, express or implied, with
  regard to Source, except that MasPar has the right to license
  Source. Source is provided to Licensee on an "AS IS" basis only.

 ********************************************************************
 ********************************************************************
 * This module: readMPPE
 *	AVS module which connects to the MasPar MPPE debugger.
 *
 * Authors:
 *	Zaide Liu
 *      Thomas D. Linton, Jr.
 ********************************************************************
 * Revision control history:
 * $Log:	readMPPE.c,v $
 * Revision 1.9  91/10/23  16:36:10  jothy
 * New copyright notice.
 * Fix socket read bug.
 * 
 * Revision 1.8  91/10/17  17:23:20  linton
 * changed license notice
 * 
 * Revision 1.7  91/10/17  15:04:44  linton
 * fix include file path
 * 
 * Revision 1.6  91/10/11  08:11:33  linton
 * fixed problem with comment
 * 
 * Revision 1.5  91/10/11  07:59:12  linton
 * clean up debug messages, remove copyright
 * 
 * Revision 1.4  91/10/10  14:46:36  linton
 * *** empty log message ***
 * 
 * Revision 1.3  91/10/05  11:21:44  linton
 * added loop to socket reads (tdl)
 * 
 * Revision 1.2  91/10/03  15:48:09  linton
 * *** empty log message ***
 * 
 * Revision 1.1  91/10/02  13:05:24  linton
 * Initial revision
 * 
 */
/*====================================================================
  includes
====================================================================*/
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <malloc.h>
#include <pwd.h>
#include <avs.h>
#include <field.h>




/*====================================================================
  typedefs and defines
====================================================================*/
#ifndef VERSION
#define VERSION "%W% Maspar %E";
#endif

#define MODULE_NAME           "readMPPE"

#define HOST_PARAM_NAME       "MPPE host"
#define VIEW_PARAM_NAME       "view directly"
#define LIST_PARAM_NAME       "Variable Files"
#define RETRIEVE_PARAM_NAME   "retrieve file"
#define DELETE_PARAM_NAME     "delete file"
#define STORAGE_PARAM_NAME    "Storage area:"

#define OUTPORT_1D        "outport 1d"
#define OUTPORT_2D        "outport 2d"
#define OUTPORT_3D        "outport 3d"
#define OUTPORT_ND        "outport Nd"
#define VAR_FILE_SUFFIX   ".var"

#ifndef _AVS_H_defined
#define AVS_TYPE_BYTE		(0)
#define AVS_TYPE_INTEGER	(1)
#define AVS_TYPE_FLOAT		(2)
#define AVS_TYPE_DOUBLE 	(3)
#endif  _AVS_H_defined
#define NONAVS_TYPE_SHORT	(4)
#define NONAVS_SUPPORT		(5)


#define PATH_NAME_LENGTH (MAXPATHLEN + 10)
#define HOST_NAME_LENGTH (256)


/************ defined in mdmsg.h */
#define AVS_HOOKUP_MSG	  (11)
#define READY_MSG         (3)
#define MASPARD           "maspard"

#define MAXIMUM_TRY	  (5)	/* number of socket connect attempts */

/************ error/success codes */
#define SUCCESS           (0)
#define ERROR_SOCKET      (1)
#define ERROR_MASPARD     (2)
#define ERROR_MPPE        (3)
#define ERROR_MALLOC      (4)
#define ERROR_DATA        (5)
#define ERROR_END_CONNECTION (6)


/************ debugging */
#ifdef YESDEBUG
#define QDBG(xxx, yyy)  fprintf(stderr,xxx,yyy);
#else
#define QDBG(xxx, yyy) ;
#endif


/*====================================================================
  Declarations
====================================================================*/
extern int Description();
static char *StrDup(char *s);
static void StoreField(AVSfield *field, char *storageDirectory,
		       char *varName);
static void ReadField(char *fileName, AVSfield **field);
static void SendField(AVSfield *field);
static void FreeField(AVSfield **field);
static void DeleteField(char *fileName);
static int tcpServer(char *mppeHost, int *port);
static int tcpClient(char *machine, char *servname);
static int ResetSockets(int *sock_maspard, int *sock_mppe);
static int *ShortToInt(short *data, int nShorts);
static int ReadSock(int socket, char *data, int nBytes);
static void HandleError(int error, int *sock_mppe, int *sock_maspard);
static int RegisterWithMaspard(char *mppeHost, int *sock_maspard, 
			       int port_serv);
static int recv_ready_msg(int sock_maspard);
static int send_avs_hookup_msg(int sock_maspard, 
			       int port, int uid, char *username);
static int ReceiveMessage(int s, AVSfield **outField, char **varName);
static FILE *OpenStorageFile (char *dir, char *name, char *suffix);
static int get_unit_size(int type);




/*====================================================================
  static variables
====================================================================*/
static char FileVersion[] = VERSION; 




/*====================================================================
  Definitions
====================================================================*/
static char *StrDup(char *s)
/*--------------------------------------------------------------------
  Allocates memory for a new string and copies s into it. Returns
  a pointer to the new string.
*/
{
  char *tmp;

  tmp = malloc(strlen(s) + 1);
  if (tmp) { strcpy(tmp, s); }
  return(tmp);
}



extern int Description(void)
/*--------------------------------------------------------------------
  Sets up widgets, etc. for this module.
*/
{
  char wdAndName[PATH_NAME_LENGTH];
  char wd[PATH_NAME_LENGTH];
  char myhost[HOST_NAME_LENGTH];
  int paramNum;

  /*---------- set module name and type */
  AVSset_module_name(MODULE_NAME, MODULE_DATA);

  /*---------- set up ports */
  AVScreate_output_port(OUTPORT_1D, "field 1D scalar uniform");  
  AVScreate_output_port(OUTPORT_2D, "field 2D scalar uniform");  
  AVScreate_output_port(OUTPORT_3D, "field 3D scalar uniform");  
  AVScreate_output_port(OUTPORT_ND, "field scalar uniform");  

  /*---------- set up control widgets */
  gethostname(myhost, sizeof(myhost));
  AVSadd_parameter(HOST_PARAM_NAME, "string", StrDup(myhost), 0, 0);
  AVSadd_parameter(VIEW_PARAM_NAME, "boolean", 0, 0, 0);
  if (!getwd(wd)) { sprintf(wd,"."); }
  sprintf(wdAndName,"%s/tmp",wd);
  paramNum = AVSadd_parameter(LIST_PARAM_NAME, "string", StrDup(wdAndName),
			      "", VAR_FILE_SUFFIX);
  AVSconnect_widget(paramNum, "browser");
  AVSadd_parameter(RETRIEVE_PARAM_NAME, "oneshot", 0, 0, 1);
  AVSadd_parameter(DELETE_PARAM_NAME, "oneshot", 0, 0, 1);

  QDBG("Description Function - 2\n","");
  paramNum = AVSadd_parameter(STORAGE_PARAM_NAME, "string", StrDup(wd), "", "");
  QDBG("Description Function - 3\n","");
}



static FILE *OpenStorageFile (char *dir, char *name, char *suffix)
/*--------------------------------------------------------------------
  Suppose 'name' is 'x' and 'suffix' is '.var'. This function
  looks in 'dir' for a file named 'x.#.var', where # is a positive
  version number. If any such files exist, then a new file with
  # one greater than the last number. Otherwise the file 
  'x.1.var' is opened. If the open fails, an AVS error message
  is written and 0 is returned; otherwise the file pointer is
  returned.
*/
{
  DIR *dp;
  struct direct *entry;
  int nLen, eLen, sLen, version, newVersion;
  char *cp, *entryName;
  char fileName[PATH_NAME_LENGTH];
  FILE *fp;

  QDBG("OpenStorageFile: dir='%s'\n",dir);
  if ((dp = opendir(dir)) == NULL) {
    AVSmessage(FileVersion, AVS_Warning, 0,"OpenStorageFile", 0,
	       "unable to open directory '%s'",dir);
    return(0);
  }

  if (cp = rindex(name, '/')) { name = ++cp; }

  nLen = strlen(name);
  sLen = strlen(suffix);
  errno = 0;
  version = 0;

  while(entry = readdir(dp)) {

    /*----- check name */
    entryName = entry->gd_name;
    if (cp = rindex(entryName, '/')) { entryName = ++cp; }
    QDBG("    checking '%s'\n",entryName);

    if (strncmp(name, entryName, nLen) == 0) {
      eLen = strlen(entryName);


      /*----- check suffix */
      if (eLen > sLen && (strcmp(&entryName[eLen - sLen], suffix) == 0)) {
	newVersion = atoi(&entryName[nLen + 1]);
	if (newVersion > version) {
	  version = newVersion;
	}
      }
      QDBG("    entryName='%s'\n",entryName);
      QDBG("    version=%d\n",version);
    }
  }

  if (errno != 0) {
    AVSmessage(FileVersion, AVS_Warning, 0,"OpenStorageFile", 0,
	       "error reading directory '%s'",dir);
    return(0);
  }
  
  version = (version > 0) ? ++version : 1;

  sprintf(fileName, "%s/%s.%d%s", dir, name, version, VAR_FILE_SUFFIX);
  QDBG("    fileName='%s'\n",fileName);
  fp = fopen(fileName,"w");

  if (!fp) {
    AVSmessage(FileVersion, AVS_Warning, 0,"OpenStorageFile", 0,
	       "unable to open file '%s'",fileName);
  }
  return(fp);
}





static void StoreField(AVSfield *inField, char *storageDirectory, 
		       char *varName)
/*--------------------------------------------------------------------
  Writes 'inField' to a file in the specified directory.
*/
{
  FILE *fp;
  int nElements, i;

  QDBG("StoreField - 1\n","");
  
  fp = OpenStorageFile(storageDirectory, varName, VAR_FILE_SUFFIX);
  if (!fp) { return; }
  
  /*---------- first the structure itself */
  fwrite(inField, sizeof(AVSfield), 1, fp);

  /*---------- now the dimensions and data */
  nElements = 1;
  for(i=0; i<inField->ndim; i++) { nElements *= (inField->dimensions)[i]; }
  fwrite(inField->dimensions, sizeof(int), inField->ndim, fp);
  fwrite(inField->field_union.field_data_char, inField->size, nElements, fp);

  fclose(fp);
}


static void FreeField(AVSfield **field)
/*--------------------------------------------------------------------
  Frees memory associated with field and sets *field to 0.
*/
{
  AVSfield *fp;

  QDBG("FreeField - begin\n","");
  if (fp = *field) {
    if (fp->dimensions) { free(fp->dimensions); }
    if (fp->field_union.field_data_char) { 
      free(fp->field_union.field_data_char);
    }
    free(fp);
    *field = 0;
  }
  QDBG("FreeField - end\n","");
}



static void SendField(AVSfield *field)
/*--------------------------------------------------------------------
  Sends 'field' to the appropriate output port, based on its
  dimension.
*/
{
  QDBG("SendField - begin\n","");
  switch(field->ndim) {
  case 1: AVScorout_output(field, 0, 0, field); break;
  case 2: AVScorout_output(0, field, 0, field); break;
  case 3: AVScorout_output(0, 0, field, field); break;
  case 4:
  case 5:
  case 6:
  case 7: AVScorout_output(0, 0, 0, field); break;
  default:
    QDBG("SendField: ndim = %d\n",field->ndim);
    AVSmessage(FileVersion, AVS_Warning, 0,"SendField", 0,
	       "unsupported number of dimensions (%d) in input",
	       field->ndim);
  }
  QDBG("SendField - end\n","");
}
  


static void ReadField(char *fileName, AVSfield **outField)
/*--------------------------------------------------------------------
  Reads 'outField' from 'fileName'.
*/
{
  FILE *fp;
  int nElements, i, type, ndim, *dim;
  AVSfield inField;
  char *data;

  QDBG("ReadField - 1\n","");

  FreeField(outField);
  if (!fileName) { return; }
  fp = fopen(fileName, "r");

  /*---------- first the structure itself */
  fread(&inField, sizeof(AVSfield), 1, fp);

  /*---------- read the dimensions and data */
  ndim = inField.ndim;
  type = inField.type;
  dim = (int *) malloc(sizeof(int) * ndim);
  fread(dim, sizeof(int), ndim, fp);

  nElements = 1;
  for(i=0; i< ndim; i++) { nElements *= dim[i]; }
  data = malloc(inField.size * nElements);
  fread(data, inField.size, nElements, fp);
  fclose(fp);

  switch(ndim) {
  case 7:
    *outField = AVSbuild_field(7, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2],
			       dim[3], dim[4], dim[5], dim[6], data); break;
  case 6:
    *outField = AVSbuild_field(6, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2],
			       dim[3], dim[4], dim[5],         data); break;
  case 5:
    *outField = AVSbuild_field(5, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2],
			       dim[3], dim[4],                 data); break;
  case 4:
    *outField = AVSbuild_field(4, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2],
			       dim[3],                         data); break;
  case 3:
    *outField = AVSbuild_field(3, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2], data); break;
  case 2:
    *outField = AVSbuild_field(2, 1, UNIFORM, 0, type, dim[0], dim[1], data); break;
  case 1:
    *outField = AVSbuild_field(1, 1, UNIFORM, 0, type, dim[0], data); break;
  default:
    AVSmessage(FileVersion, AVS_Warning, 0, "DeleteField", 0,
	       "dimension %d not yet handled\n",ndim);
  }
}


static void DeleteField(char *fileName)
/*--------------------------------------------------------------------
  Deletes 'fileName'.
*/
{
  QDBG("DeleteField - begin\n","");

  if (!fileName || strlen(fileName) < 1) { 
    AVSmessage(FileVersion, AVS_Warning, 0, "DeleteField", 0,
	       "bad file name");
    return;
  }

  if (strcmp(AVSmessage(FileVersion, AVS_Warning, 0, "DeleteField",
		 "Ok!Cancel",
		 "Confirm deletion of '%s'",fileName), "Ok") == 0) {
    if (unlink(fileName) != 0) {
      AVSmessage(FileVersion, AVS_Warning, 0,"DeleteField", 0,
		 "unable to delete file '%s'",fileName);
    }
  }
  QDBG("DeleteField - end\n","");
}




static int tcpServer(char *mppeHost, int *port)
/*--------------------------------------------------------------------
  'mppeHost' is the host to listen for. On return, 
  'port' will be set to an arbitrarily assigned port number.
   Returns a bound, listened-to socket, or -1 on error.
*/
{
  struct hostent *hp;
  struct sockaddr_in saddr;
  int slen = sizeof(saddr);
  int on = 1;
  int sock;
  
  hp = gethostbyname(mppeHost);
  bzero(&saddr, sizeof(saddr));
  saddr.sin_family = AF_INET;
  saddr.sin_port = 0;
  bcopy(hp->h_addr, &saddr.sin_addr, hp->h_length);

  sock = socket(AF_INET, SOCK_STREAM, 0);
  setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, 0, 0);

  if (bind(sock, &saddr, sizeof(saddr))) {
    perror("tcpServer() bind problem"); return -1;
  }

  if (getsockname(sock, &saddr, &slen) != 0) {
    perror("tcpServer() getsockname problem"); return -1;
  }
  *port = htons(saddr.sin_port);

  if (listen(sock, 5) != 0){
    perror("tcpServer() listen problem");  return(-1);
  }
  return(sock);
}

static int ResetSockets(int *sock_maspard, int *sock_mppe)
/*--------------------------------------------------------------------
  Shuts down sockets that will now longer be valid because
  of, e.g., a change of host name.
*/
{
  shutdown(*sock_maspard, 2);  close(*sock_maspard);  *sock_maspard = -1;
  shutdown(*sock_mppe, 2);     close(*sock_mppe);     *sock_mppe = -1;
}


static int *ShortToInt(short *data, int nShorts)
/*--------------------------------------------------------------------
  Reads 'nShorts' shorts from array 'data', casts them to ints,
  and returns a pointer to a new array holding the ints.
*/
{
  register int i;
  int *out;

  out = (int *) malloc(nShorts * sizeof(int));
  if (!out) return(0);

  for(i=0; i<nShorts; i++) {
    out[i] = (int) data[i];
  }
  return(out);
}


static int ReadSock(int socket, char *data, int nBytes)
/*--------------------------------------------------------------------
  Reads 'nBytes' from 'socket' into 'data'. Returns 1 if error
  and 0 otherwise.
*/
{
  int nread;

  QDBG("    ReadSock: nBytes = %d\n",nBytes);
 
  while ((nread = read(socket, data, nBytes)) < nBytes) {
    QDBG("    ReadSock: nread=%d\n",nread);
    if (nread > 0) {
      nBytes -= nread;
	data += nread;
    } else {
      QDBG("    ReadSock: error return\n","");
      return(1);
    }
  }
  QDBG("    ReadSock: success return\n","");
  return(0);
}




static int ReceiveMessage(int s, AVSfield **outField, char **varName)
/*--------------------------------------------------------------------
  Reads in an array of data from socket s and returns a new 
  uniform scalar field in *outField. *varName is set to the
  array variable name.
*/
{
  AVSfield *fp;
  int mid, varNameBytes, dimBytes, dataBytes;
  int type, *dim, ndim, i, nread;
  char *data, *data2;

  FreeField(outField);

  QDBG("ReceiveMessage: start","");
  nread = read(s, &mid, sizeof(int));
  if (nread == 0) {
    return ERROR_END_CONNECTION;
  } else if (nread != sizeof(int)) {
    return ERROR_DATA;
  }

  if (ReadSock(s, (char *) &type,   sizeof(int))) return ERROR_DATA;
  if (ReadSock(s, (char *) &varNameBytes, sizeof(int))) return ERROR_DATA;
  if ((*varName = malloc(varNameBytes)) == NULL)        return ERROR_MALLOC;
  if (ReadSock(s, *varName, varNameBytes))              return ERROR_DATA;
  if (ReadSock(s, (char *) &dimBytes, sizeof(int)))         return ERROR_DATA;
  if ((dim = (int *) malloc(dimBytes)) == NULL)      return ERROR_MALLOC;
  if (ReadSock(s, (char *) dim, dimBytes))     return ERROR_DATA;
  ndim = dimBytes / sizeof(int);
  if (ReadSock(s, (char *) &dataBytes, sizeof(int)))        return ERROR_DATA;
  if ((data = malloc(dataBytes)) == NULL)                   return ERROR_MALLOC;
  if (ReadSock(s, data, dataBytes))                         return ERROR_DATA;

  QDBG("ReceiveMessage: mid=%d\n",mid);
  QDBG("ReceiveMessage: type=%d\n",type);
  QDBG("ReceiveMessage: varNameBytes=%d\n",varNameBytes);
  QDBG("ReceiveMessage: varName='%s'\n",*varName);
  QDBG("ReceiveMessage: dimBytes=%d\n",dimBytes);
#ifdef YESDEBUG
  for(i=0; i<dimBytes/sizeof(int); i++) {
    fprintf(stderr,"Receive Message: dim[%d]=%d\n",i,dim[i]);
  }
#endif
  QDBG("ReceiveMessage: dataBytes=%d\n",dataBytes);
#ifdef YESDEBUG    
 { float *f = (float *) data;
  for(i=0; i<10; i++) {
    fprintf(stderr,"Receive Message: data[%d]=%e  %d\n",i, f[i], (int) f[i]);
  }
 }
#endif


  if (type == NONAVS_TYPE_SHORT) {
    type = AVS_TYPE_INTEGER;
    data2 = (char *) ShortToInt((short *)data, dataBytes / sizeof(short));
    free(data); data = data2;
    if (!data) return(ERROR_DATA);
  }

  switch(ndim) {
  case 7:
    *outField = AVSbuild_field(7, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2],
			       dim[3], dim[4], dim[5], dim[6], data); break;
  case 6:
    *outField = AVSbuild_field(6, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2],
			       dim[3], dim[4], dim[5],         data); break;
  case 5:
    *outField = AVSbuild_field(5, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2],
			       dim[3], dim[4],                 data); break;
  case 4:
    *outField = AVSbuild_field(4, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2],
			       dim[3],                         data); break;
  case 3:
    *outField = AVSbuild_field(3, 1, UNIFORM, 0, type, dim[0], dim[1], dim[2], data); break;
  case 2:
    *outField = AVSbuild_field(2, 1, UNIFORM, 0, type, dim[0], dim[1], data); break;
  case 1:
    *outField = AVSbuild_field(1, 1, UNIFORM, 0, type, dim[0], data); break;
  default:
    fprintf(stderr,"dimension %d not yet handled\n",ndim);
    return(ERROR_DATA);
  }
  return(SUCCESS);
}


static int get_unit_size(int type)
/*--------------------------------------------------------------------
  Return the size of a given AVS type.
*/
{
  switch(type) {
  case AVS_TYPE_BYTE: return(sizeof(char));
  case AVS_TYPE_INTEGER:  return(sizeof(int));
  case AVS_TYPE_REAL:     return(sizeof(float));
  case AVS_TYPE_DOUBLE:   return(sizeof(double));
  case NONAVS_TYPE_SHORT: return(sizeof(short));
  default:
    return(sizeof(int));
  }
}
    


static void HandleError(int error, int *sock_mppe, int *sock_maspard)
/*--------------------------------------------------------------------
  Treats predefined error codes using AVS messages and other cleanup
  actions as necessary.
*/
{
  QDBG("HandleError: error = %d\n",error);
  switch(error) {
  case ERROR_MPPE:
    shutdown(*sock_mppe, 2); 
    close(*sock_mppe);  
    *sock_mppe = -1;
    AVSmessage(FileVersion, AVS_Warning, 0, "HandleError", 0,
	       "mppe socket problem");
    break;
  case ERROR_MASPARD:
    shutdown(*sock_maspard, 2);
    close(*sock_maspard);
    *sock_maspard = -1;
    AVSmessage(FileVersion, AVS_Warning, 0, "HandleError", 0,
	       "maspard socket problem");
    break;
  case ERROR_MALLOC:
    AVSmessage(FileVersion, AVS_Warning, 0, "HandleError", 0,
	       "malloc failed");
    break;
  case ERROR_DATA:
    shutdown(*sock_mppe, 2); 
    close(*sock_mppe);  
    *sock_mppe = -1;
    AVSmessage(FileVersion, AVS_Warning, 0, "HandleError", 0,
	       "error reading data from mppe");
    break;
  case ERROR_END_CONNECTION:
    shutdown(*sock_mppe, 2); 
    close(*sock_mppe);  
    *sock_mppe = -1;
    break;
  case ERROR_SOCKET:
    AVSmessage(FileVersion, AVS_Warning, 0, "HandleError", 0,
	       "socket error");
    break;
  }
}


static int RegisterWithMaspard(char *mppeHost, int *sock_maspard, 
			       int port_serv)
/*--------------------------------------------------------------------
  Connects a socket to maspard, then sends it an 'avs_hookup' message
  and waits for a 'recv_ready' message. This lets maspard know that
  that there's an avs reader (this program) listening for data.
*/
{
  int myuid, error;
  char *myname;
  struct passwd *pwd;
  
  *sock_maspard = tcpClient(mppeHost, MASPARD);
  if (*sock_maspard <= 0)  return ERROR_MASPARD;
  
  myuid = (int) getuid();  
  pwd = getpwuid(myuid);
  myname = pwd->pw_name;
  error = send_avs_hookup_msg(*sock_maspard, port_serv, myuid, myname);
  if (error) return ERROR_MASPARD;
  recv_ready_msg(*sock_maspard);
}



static int tcpClient(char *machine, char *servname)
/*--------------------------------------------------------------------
  Connects to a server socket with on 'machine' with service
  name 'servname'. Returns the socket if successful or
  -1 if not.
*/
{
  int i;
  int sock;
  char hostname[HOST_NAME_LENGTH];
  struct hostent *hp;
  struct sockaddr_in saddr;
  struct servent *sp;
  
  if (!servname) return -1;
  sp = getservbyname(servname, "tcp");
  if (!sp) {
    perror("getservbyname");  return -1;
  }
  
  if (!machine) {
    gethostname(hostname, sizeof(hostname));
  } else {
    strcpy(hostname, machine);
  }
  
  hp = gethostbyname(hostname);
  if (!hp) {
    perror("gethostbyname");  return -1;
  }
  
  bzero(&saddr, sizeof(saddr));
  saddr.sin_family = AF_INET;
  saddr.sin_port = sp->s_port;
  bcopy(hp->h_addr, &saddr.sin_addr, hp->h_length);
  
  for (i=0; i<MAXIMUM_TRY; i++) {
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock == -1) {
      perror("socket creation"); return -1;
    }
    
    if (connect(sock, &saddr, sizeof(saddr)) == 0) {
      return(sock);
    } else {
      close(sock);
    }
    sleep(1);
  }

  /* we have tried many times, but failed */
  return(-1);
}


static int recv_ready_msg(int sock_maspard)
/*--------------------------------------------------------------------
  Get 'ready' message back from the maspard socket.
*/
{
  int in_type;
  read(sock_maspard, (char *)&in_type, sizeof(int));
  if (ntohl(in_type) != READY_MSG) {
    return ERROR_MASPARD;
  } else {
    return SUCCESS;
  }
}



static int send_avs_hookup_msg(int sock_maspard, 
			       int port, int uid, char *username)
/*--------------------------------------------------------------------
  Sends an 'avs_hookup' message on the maspard socket.
*/
{
  int out_type = htonl(AVS_HOOKUP_MSG);
  int out_port = htonl(port);
  int out_uid  = htonl(uid);
  errno = 0;
  write(sock_maspard, (char *)&out_type, sizeof(int));
  write(sock_maspard, (char *)&out_port, sizeof(int));
  write(sock_maspard, (char *)&out_uid,  sizeof(int));
  /* write(sock_maspard, username, strlen(username) + 1);
  */
  if (errno) {
    return ERROR_MASPARD;
  } else {
    return SUCCESS;
  }
}

  

/*====================================================================
  Main
====================================================================*/

main(int argc, char **argv)
{
  AVSfield *data;		    /* data received from socket */
  char mppeHost[HOST_NAME_LENGTH];  /* name for mppe host */
  int rc;			/* return code */
  char *varName;		/* variable name */

  /*---------- communication data structures */
  int sock_serv    = -1;	/* server listener socket */
  int port_serv    = -1;	/* server listener port */
  int sock_mppe    = -1;	/* data connection to mppe */
  int sock_maspard = -1;	/* connection to maspard */
  int socketError;		/* 1 = error, 0 = ok */
  int newData = 0;		/* 1 = new data available */
  struct timeval timeout;
  fd_set rmask;			/* mask for select */

  /*---------- AVS input parameters */
  char *newmppeHost;		/* new name for mppe host */
  int viewParam;		/* 1 for direct data view mode */
  char *listParam;		/* name of selected var file */
  char *storageParam;		/* name of storage directory */
  int retrieveParam;		/* 1 if user wants to view file var */
  int deleteParam;		/* 1 if user wants to delete file var */



  
  /*-------------------- initialization */
  AVScorout_init(argc, argv, Description);
  timeout.tv_sec = 1;  timeout.tv_usec = 0;
  data = 0;
  gethostname(mppeHost, sizeof(mppeHost));
  
  sock_serv = tcpServer(mppeHost, &port_serv);
  if (sock_serv <= 0) {
    HandleError(ERROR_SOCKET, &sock_mppe, &sock_maspard);
  }
  

  /*-------------------- main loop */
  for(;;) {
    socketError = 0;
    AVScorout_input(&newmppeHost, &viewParam, &listParam,
		    &retrieveParam, &deleteParam, &storageParam);

    /*---------- poll to check if user wants a new host name */
    if (strcmp(mppeHost, newmppeHost) != 0) {
      ResetSockets(&sock_maspard, &sock_mppe);
      strcpy(mppeHost, newmppeHost);
      rc = RegisterWithMaspard(mppeHost, &sock_maspard, port_serv);
      if (rc == SUCCESS) {
	socketError = 0;
      } else {
	HandleError(rc, &sock_mppe, &sock_maspard);
	socketError = 1;
      }
    }

    /*---------- poll to establish mppeback connection */
    if (!socketError) {
      if (sock_mppe <= 0) {
	FD_ZERO(&rmask);  FD_SET(sock_serv, &rmask);
	if (select(sock_serv+1, &rmask, 0, 0, &timeout) <= 0) {
	  socketError = 1;
	} else {
	  sock_mppe = accept(sock_serv, NULL, NULL);
	}
      }
    }
      
    /*---------- poll for data from mppeback */
    if (!socketError) {
      if (sock_mppe <= 0) {
	socketError = 1; 
      } else {
	FD_ZERO(&rmask);  FD_SET(sock_mppe, &rmask);
	if (select(sock_mppe+1, &rmask, 0, 0, &timeout) > 0) {
	  rc = ReceiveMessage(sock_mppe, &data, &varName);
	  if (rc != SUCCESS) {
	    HandleError(rc, &sock_mppe, &sock_maspard);
	  } else {
	    newData = 1;
	  }
	}
      }
    }
      

    /*---------- route data if any has arrived */
    if (data && newData) {
      newData = 0;
      if (viewParam) {
	SendField(data);
      } else {
	StoreField(data, storageParam, varName);
      }
    }

    /*---------- deal with other commands */
    if (!viewParam) {
      if (retrieveParam) {
	ReadField(listParam, &data);
	SendField(data);
	AVSmodify_parameter(RETRIEVE_PARAM_NAME, AVS_VALUE, 0, 0, 0);
      }
      if (deleteParam) {
	DeleteField(listParam);
	AVSmodify_parameter(DELETE_PARAM_NAME, AVS_VALUE, 0, 0, 0);
      }
    }
  }
}
